import typing
import datetime
import json

TWITCH_API_TIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"

def str_to_datetime(value : str) -> datetime.datetime:
  try:
    return datetime.datetime.strptime(value, TWITCH_API_TIME_FORMAT).replace(tzinfo=datetime.timezone.utc)
  except:
    return None

class LusciousBaseObject:
  class DoesNotExist(Exception):
    def __init__(self, message : str):
      super().__init__(message)
  
  class InvalidFormat(KeyError):
    def __init__(self, message : str):
      super().__init__(message)
      
  @property
  def attributes(self) -> typing.List[str]:
    return []
  
  @property
  def optional_attributes(self) -> typing.List[str]:
    return []
  
  def parse_attribute(self, key : str, value : object) -> bool:
    setattr(self, key, value)
    return True
    
  def __init__(self, object_data : dict):
    missing_keys = []
    invalid_keys = []
    
    for key in self.attributes:
      if key not in object_data:
        missing_keys.append(key)
        continue
      
      success = self.parse_attribute(key, object_data.get(key, None))
      
      if not success:
        invalid_keys.append(key)
        continue
    
    for key in self.optional_attributes:
      self.parse_attribute(key, object_data.get(key, None))
        
    if len(missing_keys) > 0:
      raise LusciousBaseObject.InvalidFormat(f"{type(self).__name__} JSON data is missing keys: {", ".join(missing_keys)}")
    
    if len(invalid_keys) > 0:
      raise LusciousBaseObject.InvalidFormat(f"{type(self).__name__} JSON data has invalid values for keys: {", ".join(invalid_keys)}")

class TwitchClip(LusciousBaseObject):
  clip_id : str = None
  url : str = None
  embed_url : str = None
  broadcaster_id : str = None
  broadcaster_name : str = None
  creator_id : str = None
  creator_name : str = None
  video_id : str = None
  game_id : str = None
  language : str = None
  title : str = None
  view_count : int = None
  created_at : datetime.datetime = None
  thumbnail_url : str = None
  duration : float = None
  vod_offset : int = None
  is_featured : bool = None
  
  def __str__(self) -> str:
    return f"{self.title} ({self.clip_id})"
      
  @property
  def attributes(self) -> typing.List[str]:
    return [
      "id",
      "url",
      "embed_url",
      "broadcaster_id",
      "broadcaster_name",
      "creator_id",
      "creator_name",
      "video_id",
      "game_id",
      "language",
      "title",
      "view_count",
      "created_at",
      "thumbnail_url",
      "duration",
      "vod_offset",
      "is_featured",
    ]
  
  def parse_attribute(self, key : str, value : object) -> bool:
    if key == "id":
      setattr(self, "clip_id", value)
    elif key == "created_at":
      try:
        setattr(self, key, datetime.datetime.strptime(value, TWITCH_API_TIME_FORMAT))
      except:
        return False
    else:
      setattr(self, key, value)
      
    return True
  
class GQL_PlaybackAccessToken(LusciousBaseObject):
  signature : str = None
  value : dict = None
  
  @property
  def attributes(self) -> typing.List[str]:
    return [
      "signature",
      "value",
    ]
    
  def parse_attribute(self, key : str, value : object):
    if key == "value":
      setattr(self, key, json.loads(value))
    else:
      setattr(self, key, value)
    return True
    
class GQL_VideoQuality(LusciousBaseObject):
  frameRate : float = -1.0
  quality : str = None
  sourceURL : str = None
    
  def __str__(self):
    return f"{self.quality} @ {self.frameRate:.2f}Hz: {self.sourceURL}"
  
  @property
  def attributes(self) -> typing.List[str]:
    return [
      "frameRate",
      "quality",
      "sourceURL",
    ]
  
class GQL_Clip(LusciousBaseObject):
  clipId : str = None
  playbackAccessToken : GQL_PlaybackAccessToken = None
  videoQualities : typing.List[GQL_VideoQuality] = []
  
  @property
  def attributes(self) -> typing.List[str]:
    return [
      "id",
      "playbackAccessToken",
      "videoQualities",
    ]
    
  def parse_attribute(self, key : str, value : object) -> bool:
    if key == "id":
      setattr(self, "clipId", value)
    elif key == "playbackAccessToken":
      setattr(self, key, GQL_PlaybackAccessToken(value))
    elif key == "videoQualities":
      setattr(self, key, [GQL_VideoQuality(vq) for vq in value])
    else:
      setattr(self, key, value)
    return True

class GQL_Video(LusciousBaseObject):
  videoId : str = None
  title : str = None
  thumbnailURLs : typing.List[str] = []
  createdAt : datetime.datetime = None
  publishedAt : datetime.datetime = None
  lengthSeconds : int = -1
  owner : typing.Dict[str, str] = {}
  extensions : typing.Dict[str, typing.Union[str, int]] = {}
  
  @property
  def attributes(self) -> typing.List[str]:
    return [
      "id",
      "title",
      "thumbnailURLs",
      "createdAt",
      "publishedAt",
      "lengthSeconds",
      "owner",
    ]
    
  @property
  def optional_attributes(self) -> typing.List[str]:
    return [ "extensions" ]
    
  def parse_attribute(self, key : str, value : object) -> bool:
    if key == "id":
      setattr(self, "videoId", value)
    elif key == "createdAt" or key == "publishedAt":
      dt = str_to_datetime(value)
      setattr(self, key, dt)
      if dt is None:
        return False
    else:
      setattr(self, key, value)
      
    return True
      

class CreateClipResponse(LusciousBaseObject):
  edit_url : str = None
  clip_id : str = None
  
  @property
  def attributes(self) -> typing.List[str]:
    return [
      "edit_url",
      "id",
    ]
    
  def parse_attribute(self, key : str, value : object) -> bool:
    if key == "id":
      setattr(self, "clip_id", value)
    else:
      setattr(self, key, value)
      
    return True
    
class TwitchVideo(LusciousBaseObject):
  video_id : str = None
  stream_id : str = None
  user_id : str = None
  user_login : str = None
  user_name : str = None
  title : str = None
  description : str = None
  created_at : datetime.datetime = None
  published_at : datetime.datetime = None
  url : str = None
  thumbnail_url : str = None
  viewable : bool = None
  view_count : int = None
  language : str = None
  type : str = None
  duration : str = None
  muted_segments : typing.Optional[typing.List[typing.Dict[str, int]]] = None
  
  def __str__(self) -> str:
    return f"{self.title} ({self.video_id})"
      
  @property
  def attributes(self) -> typing.List[str]:
    return [
      "id",
      "stream_id",
      "user_id",
      "user_login",
      "user_name",
      "title",
      "description",
      "created_at",
      "published_at",
      "url",
      "thumbnail_url",
      "viewable",
      "view_count",
      "language",
      "type",
      "duration",
      "muted_segments"
    ]
  
  def parse_attribute(self, key : str, value : object) -> bool:
    if key == "id":
      setattr(self, "video_id", value)
    elif key == "created_at" or key == "published_at":
      dt = str_to_datetime(value)
      setattr(self, key, dt)
      if dt is None:
        return False
    else:
      setattr(self, key, value)
      
    return True

class TwitchStream(LusciousBaseObject):
  stream_id : str = None
  user_id : str = None
  user_login : str = None
  user_name : str = None
  game_id : str = None
  game_name : str = None
  type : str = None
  title : str = None
  tags : typing.List[str] = []
  viewer_count : int = -1
  started_at : datetime.datetime = None
  language : str = None
  thumbnail_url : str = None
  tag_ids : typing.List[str] = []
  is_mature : bool = False
  
  def __str__(self) -> str:
    return f"{self.title} ({self.stream_id})"
  
  @property
  def attributes(self) -> typing.List[str]:
    return [
      "id",
      "user_id",
      "user_login",
      "user_name",
      "game_id",
      "game_name",
      "type",
      "title",
      "tags",
      "viewer_count",
      "started_at",
      "language",
      "thumbnail_url",
      "is_mature",
    ]
    
  @property
  def optional_attributes(self) -> typing.List[str]:
    return [ "tag_ids" ]
    
  def parse_attribute(self, key, value):
    if key == "id":
      setattr(self, "stream_id", value)
    elif key == "started_at":
      dt = str_to_datetime(value)
      setattr(self, key, dt)
      if dt is None:
        return False
    else:
      setattr(self, key, value)
      
    return True
      
class TwitchUser(LusciousBaseObject):
  user_id : str = None
  login : str = None
  display_name = None
  user_type : str = None
  broadcaster_type : str = None
  description : str = None
  profile_image_url : str = None
  offline_image_url : str = None
  view_count : int = None
  email : str = None
  created_at : datetime.datetime = None
  
  def __str__(self) -> str:
    return f"{self.display_name}/{self.login} ({self.user_id})"
  
  @property
  def attributes(self) -> typing.List[str]:
    return [
      "id",
      "login",
      "display_name",
      "type",
      "broadcaster_type",
      "description",
      "profile_image_url",
      "offline_image_url",
      "created_at",
    ]
    
  @property
  def optional_attributes(self) -> typing.List[str]:
    return [
      "view_count",
      "email",
    ]
    
  def parse_attribute(self, key, value):
    if key == "id":
      setattr(self, "user_id", value)
    elif key == "created_at":
      dt = str_to_datetime(value)
      setattr(self, key, dt)
      if dt is None:
        return False
    else:
      setattr(self, key, value)
      
    return True
  
class TwitchChannelInfo(LusciousBaseObject):
  broadcaster_id : str = None
  broadcaster_login : str = None
  broadcaster_name : str = None
  broadcaster_language : str = None
  game_name : str = None
  game_id : str = None
  title : str = None
  delay : int = None
  tags : typing.List[str] = None
  content_classification_labels : typing.List[str] = None
  is_branded_content : bool = False
  
  def __str__(self) -> str:
    return f"{self.title} - {self.broadcaster_name}"
  
  @property
  def attributes(self) -> typing.List[str]:
    return [
      "broadcaster_id",
      "broadcaster_login",
      "broadcaster_name",
      "broadcaster_language",
      "game_name",
      "game_id",
      "title",
      "delay",
      "tags",
      "content_classification_labels",
      "is_branded_content",
    ]
    
class TwitchCategoryInfo(LusciousBaseObject):
  category_id : str = None
  name : str = None
  box_art_url = None
  igdb_id = None
  
  def __str__(self) -> str:
    return f"{self.name} ({self.category_id})"
  
  @property
  def attributes(self) -> typing.List[str]:
    return [
      "id",
      "name",
      "box_art_url",
      "igdb_id",
    ]
    
  def parse_attribute(self, key, value):
    if key == "id":
      setattr(self, "category_id", value)
    else:
      setattr(self, key, value)
      
    return True
  
class TwitchEmote(LusciousBaseObject):
  emote_id : str = None
  name : str = None
  images : typing.Dict[str, str] = {}
  format : typing.List[str] = []
  scale : typing.List[str] = []
  theme_mode : typing.List[str] = []
  
  def __str__(self):
    return self.name
  
  @property
  def attributes(self) -> typing.List[str]:
    return [
      "id",
      "name",
      "images",
      "format",
      "scale",
      "theme_mode",
    ]
    
  def parse_attribute(self, key, value):
    if key == "id":
      setattr(self, "emote_id", value)
    else:
      setattr(self, key, value)
      
    return True
    
class TwitchChannelEmote(TwitchEmote):
  tier : str = None
  emote_type : str = None
  emote_set_id : str = None
  
  @property
  def attributes(self) -> typing.List[str]:
    attr_list = []
    attr_list.extend(super().attributes)
    attr_list.extend([
      "tier",
      "emote_type",
      "emote_set_id",
    ])
    return attr_list


from .lushrequests import *
from .lushwebsocket import *
from .lushevents import *
from .lushtypes import *

import typing

class TwitchAPI:
  API_URL = "https://api.twitch.tv/helix"
  OAUTH_URL = "https://id.twitch.tv/oauth2"
  AUTHENTICATED = False
  CLIENT_ID = ""
  CLIENT_SECRET = ""
  ACCESS_TOKEN = ""
  REFRESH_TOKEN = ""
  AUTH_TYPE = "client_credentials"
  EXPIRES_IN = -1
  DEFAULT_HEADERS = {}
  rlrequests = RateLimitedRequests(400, 60)
  TWITCH_WEBSOCKET = None
  
  def __add_parameters(self, url : str, params : dict):
    """Add parameters to an API request URL.

    Args:
        url (string): API endpoint
        params (string): Dictionary of parameters. See function descriptions for valid params.

    Returns:
        string: URL with params.
    """
    separator = "?"
    for k, v in params.items():
      url += f"{separator}{k}={v}"
      separator = "&"
    return url
  
  def __raise_req_error(self, response : requests.Response):
    raise Exception(f'Status {response.status_code}: {response.reason}')

  def __init__(self, credentials : dict, override_api_url : str = "", override_oauth_url : str = ""):
    """Constructor for TwitchAPI. Must pass in credentials in the form of a dictionary.

    Args:
        credentials (dict): API Credentials. "CLIENT_ID" and "CLIENT_SECRET" should be in the dict.
    """
    if override_api_url != "":
      self.API_URL = override_api_url
      
    if override_oauth_url != "":
      self.OAUTH_URL = override_oauth_url
    
    self.CLIENT_ID = credentials.get("CLIENT_ID", "")
    self.CLIENT_SECRET = credentials.get("CLIENT_SECRET", "")
    
    if self.CLIENT_ID == "":
      raise Exception("Missing Client ID.")
    
    if 'ACCESS_TOKEN' in credentials:
      self.ACCESS_TOKEN = credentials.get("ACCESS_TOKEN", "")
      self.EXPIRES_IN = credentials.get("EXPIRES_IN", 0)
      self.REFRESH_TOKEN = credentials.get("REFRESH_TOKEN", "")
      self.AUTH_TYPE = credentials.get("AUTH_TYPE", "authorization_code")
    else:
      self.AUTH_TYPE = "client_credentials"
      
      r = requests.post(f'{self.OAUTH_URL}/token?client_id={self.CLIENT_ID}&client_secret={self.CLIENT_SECRET}&grant_type=client_credentials', headers = {'Content-Type': 'application/x-www-form-urlencoded'})
    
      try:
        self.ACCESS_TOKEN = r.json()['access_token']
        self.EXPIRES_IN = r.json()["expires_in"]
      except:
        raise Exception("Failed to create access token. Invalid credentials.")
    
    self.AUTHENTICATED = True
    self.DEFAULT_HEADERS = { "Authorization": f"Bearer {self.ACCESS_TOKEN}", "Client-Id": self.CLIENT_ID }
    
  def refresh_authorization(self) -> bool:
    if self.AUTH_TYPE == "client_credentials":
      return self._refresh_client_credential_authentication()
    else:
      return self.refresh_access_token()
    
  def refresh_access_token(self, refresh_token : str = None) -> bool:
    if refresh_token == None:
      refresh_token = self.REFRESH_TOKEN
    
    url = f'{self.OAUTH_URL}/token'
    post_params = {
      'client_id': self.CLIENT_ID,
      'client_secret': self.CLIENT_SECRET,
      'grant_type': 'refresh_token',
      'refresh_token': refresh_token
    }
    url = self.__add_parameters(url, post_params)
    
    r : requests.Response = self.rlrequests.post(url = url, headers = { 'Content-Type': 'application/x-www-form-urlencoded' })
    
    try:
      self.ACCESS_TOKEN = r.json()['access_token']
      self.REFRESH_TOKEN = r.json()['refresh_token']
    except:
      self.AUTHENTICATED = False
      raise Exception(f"{r.status_code}: {r.content}.\nFailed to refresh access token.")
    
    self.DEFAULT_HEADERS = { "Authorization": f"Bearer {self.ACCESS_TOKEN}", "Client-Id": self.CLIENT_ID }
    self.AUTHENTICATED = True
    
    return self.AUTHENTICATED
  
  def _refresh_client_credential_authentication(self) -> bool:
    r = requests.post(f'{self.OAUTH_URL}/token?client_id={self.CLIENT_ID}&client_secret={self.CLIENT_SECRET}&grant_type=client_credentials', headers = {'Content-Type': 'application/x-www-form-urlencoded'})
  
    try:
      self.ACCESS_TOKEN = r.json()['access_token']
      self.EXPIRES_IN = r.json()["expires_in"]
    except:
      self.AUTHENTICATED = False
      raise Exception("Failed to create access token. Invalid credentials.")
    
    self.DEFAULT_HEADERS = { "Authorization": f"Bearer {self.ACCESS_TOKEN}", "Client-Id": self.CLIENT_ID }
    self.AUTHENTICATED = True
    
    return self.AUTHENTICATED
    

  def get_user_id(self, login : str = "") -> str:
    """Get user ID from username.

    Args:
        username (string): Username

    Returns:
        string: User ID
    """
    twitch_user = self.get_user_by_login(login)
    return twitch_user.user_id
      
  def get_user(self, user_id : str) -> TwitchUser:
    """Get user by user ID

    Args:
        user_id (str): user ID

    Returns:
        dict: user information
    """
    url = f"{self.API_URL}/users?id={user_id}"
    r : requests.Response = self.rlrequests.get(url = url, headers = self.DEFAULT_HEADERS)
    
    if r.status_code == 200:
      response_json : dict = r.json()
      response_data : typing.List[dict] = response_json.get("data", [])
      if len(response_data) > 0:
        return TwitchUser(response_data[0])
      else:
        raise TwitchUser.DoesNotExist(f"User ID \"{user_id}\" was not found.")
    elif r.status_code == 401:
      if self.refresh_authorization():
        return self.get_user(user_id)
      else:
        self.__raise_req_error(r)
    else:
      self.__raise_req_error(r)
      
  def get_user_by_login(self, login : str) -> TwitchUser:
    """Get user ID from login.

    Args:
        login (string): Login

    Returns:
        string: User ID
    """
    url = f"{self.API_URL}/users?login={login}"
    r : requests.Response = self.rlrequests.get(url = url, headers = self.DEFAULT_HEADERS)
    
    if r.status_code == 200:
      response_json : dict = r.json()
      response_data : typing.List[dict] = response_json.get("data", [])
      if len(response_data) > 0:
        return TwitchUser(response_data[0])
      else:
        raise TwitchUser.DoesNotExist(f"User login \"{login}\" was not found.")
    elif r.status_code == 401:
      if self.refresh_authorization():
        return self.get_user_by_login(login)
      else:
        self.__raise_req_error(r)
    else:
      self.__raise_req_error(r)
  
  def get_channel_info(self, broadcaster_id : str) -> TwitchChannelInfo:
    """Get Channel Information.

    Args:
        user_id (string): User ID

    Returns:
        dict: Channel information
    """
    url = f"{self.API_URL}/channels?broadcaster_id={broadcaster_id}"
    r : requests.Response = self.rlrequests.get(url = url, headers = self.DEFAULT_HEADERS)
    
    if r.status_code == 200:
      response_json : dict = r.json()
      response_data : typing.List[dict] = response_json.get("data", [])
      if len(response_data) > 0:
        return TwitchChannelInfo(response_data[0])
      else:
        raise TwitchChannelInfo.DoesNotExist(f"Broadcaster ID \"{broadcaster_id}\" was not found.")
    elif r.status_code == 401:
      if self.refresh_authorization():
        return self.get_channel_info(broadcaster_id)
      else:
        self.__raise_req_error(r)
    else:
      self.__raise_req_error(r)
      
  def get_category_by_id(self, category_id : str) -> TwitchCategoryInfo:
    """Get category information from category name

    Args:
        category_id (str): Category ID

    Returns:
        dict: Category information
    """
    url = f"{self.API_URL}/games?id={category_id}"
    r : requests.Response = self.rlrequests.get(url = url, headers = self.DEFAULT_HEADERS)
    
    if r.status_code == 200:
      response_json : dict = r.json()
      response_data : typing.List[dict] = response_json.get("data", [])
      if len(response_data) > 0:
        return TwitchCategoryInfo(response_data[0])
      else:
        raise TwitchCategoryInfo.DoesNotExist(f"Category ID \"{category_id}\" was not found.")
    elif r.status_code == 401:
      if self.refresh_authorization():
        return self.get_category_by_id(category_id)
      else:
        self.__raise_req_error(r)
    else:
      self.__raise_req_error(r)
      
  def get_category_by_name(self, category_name : str) -> TwitchCategoryInfo:
    """Get category information from category name

    Args:
        category_id (str): Category ID

    Returns:
        dict: Category information
    """
    url = f"{self.API_URL}/games?name={category_name}"  
    r : requests.Response = self.rlrequests.get(url = url, headers = self.DEFAULT_HEADERS)
    
    if r.status_code == 200:
      response_json : dict = r.json()
      response_data : typing.List[dict] = response_json.get("data", [])
      if len(response_data) > 0:
        return TwitchCategoryInfo(response_data[0])
      else:
        raise TwitchCategoryInfo.DoesNotExist(f"Category \"{category_name}\" was not found.")
    elif r.status_code == 401:
      if self.refresh_authorization():
        return self.get_category_by_name(category_name)
      else:
        self.__raise_req_error(r)
    else:
      self.__raise_req_error(r)
      
  def get_category_id(self, category_name : str) -> str:
    """Get category ID

    Args:
        category_name (str): _description_

    Returns:
        str: _description_
    """
    return self.get_category_by_name(category_name).category_id

  def get_clip(self, clip_id : str) -> TwitchClip:
    """Get info for one clip from ID.

    Args:
        clip_id (string): Video ID

    Returns:
        dict: clip info
    """
    url = f"{self.API_URL}/clips?id={clip_id}"
    r: requests.Response = self.rlrequests.get(url = url, headers = self.DEFAULT_HEADERS)
    
    if r.status_code == 200:
      response_json : dict = r.json()
      response_data : typing.List[dict] = response_json.get("data", [])
      if len(response_data) > 0:
        return TwitchClip(response_data[0])
      else:
        raise TwitchClip.DoesNotExist(f"Clip ID \"{clip_id}\" was not found.")
    elif r.status_code == 401:
      if self.refresh_authorization():
        return self.get_clip(clip_id)
      else:
        self.__raise_req_error(r)
    else:
      self.__raise_req_error(r)
      
  def create_clip(self, broadcaster_id : str, has_delay : bool = False) -> dict:
    """Create a clip of a broadcast.

    Args:
        broadcaster_id (str): Broadcaster ID
        has_delay (bool, optional): Add delay to clip. Defaults to False.

    Returns:
        dict: Dictionary containing id and edit_url
    """
    url = f'{self.API_URL}/clips'
    url = self.__add_parameters(url, { 'broadcaster_id': broadcaster_id, 'has_delay': 'true' if has_delay else 'false'})
    
    r: requests.Response = self.rlrequests.post(url = url, headers = self.DEFAULT_HEADERS)
    
    if r.status_code == 202:
      response_json : dict = r.json()
      response_data : typing.List[dict] = response_json.get("data", [])
      if len(response_data) > 0:
        return CreateClipResponse(response_data[0])
      else:
        raise CreateClipResponse.DoesNotExist("Create clip failed.")
    elif r.status_code == 401:
      if self.refresh_authorization():
        return self.create_clip(broadcaster_id, has_delay)
      else:
        self.__raise_req_error(r)
    else:
      self.__raise_req_error(r)

  def get_clips(self, params : dict) -> typing.List[TwitchClip]:
    """Get clips based on params.

    Args:
        params (dict): Dictionary of parameters for the API request. The valid params are:
            id (string): Clip ID
            broadcaster_id (string): Broadcaster ID
            game_id (string): Game/Category ID
            started_at (string): RFC3339 format, use TWITCH_API_TIME_FORMAT from this library
            ended_at (string): RFC3339 format, use TWITCH_API_TIME_FORMAT from this library
            first (int): fetch the first n clips
            before (string): reverse pagination
            after (string): forward pagination

    Returns:
        list: list of clip info
    """
    url = f"{self.API_URL}/clips"
    url = self.__add_parameters(url, params)
    
    r : requests.Response = self.rlrequests.get(url = url, headers=self.DEFAULT_HEADERS)
    
    if r.status_code == 200:
      response_json : dict = r.json()
      response_data : typing.List[dict] = response_json.get("data", [])
      response_pagination : dict = response_json.get("pagination", {})
      
      video_list = [TwitchClip(data) for data in response_data]
      cursor = response_pagination['cursor'] if "cursor" in response_pagination else ""
      
      return video_list, cursor
    elif r.status_code == 401:
      if self.refresh_authorization():
        return self.get_clips(params)
      else:
        self.__raise_req_error(r)
    else:
      self.__raise_req_error(r)

  def get_all_clips(self, params : dict) -> typing.List[TwitchClip]:
    """Get all clips based on params (auto-pagination).

    Args:
        params (dict): Dictionary of parameters for the API request. The valid params are:
            id (string): Clip ID
            broadcaster_id (string): Broadcaster ID
            game_id (string): Game/Category ID
            started_at (string): RFC3339 format, use TWITCH_API_TIME_FORMAT from this library
            ended_at (string): RFC3339 format, use TWITCH_API_TIME_FORMAT from this library
            first (int): fetch the first n clips
            before (string): reverse pagination
            after (string): forward pagination

    Returns:
        list: list of clip info
    """
    all_clips = []
    while True:
      clips, cursor = self.get_clips(params)

      for clip in clips:
        all_clips.append(clip)
      
      if cursor == "":
        return all_clips
      else:
        params["after"] = cursor

  def get_video(self, video_id : typing.Union[str, int]) -> TwitchVideo:
    """Get info for one video from ID.

    Args:
        video_id (string|int): Video ID

    Returns:
        dict: Video info
    """
    url = f"{self.API_URL}/videos?id={video_id}"
    r : requests.Response = self.rlrequests.get(url = url, headers = self.DEFAULT_HEADERS)
    
    if r.status_code == 200:
      response_json : dict = r.json()
      response_data : typing.List[dict] = response_json.get("data", [])
      if len(response_data) > 0:
        return TwitchVideo(response_data[0])
      else:
        raise TwitchVideo.DoesNotExist(f"Video ID \"{video_id}\" returned no results.")
    elif r.status_code == 401:
      if self.refresh_authorization():
        return self.get_video(video_id)
      else:
        self.__raise_req_error(r)
    else:
      self.__raise_req_error(r)

  def get_videos(self, params : dict) -> typing.Tuple[typing.List[TwitchVideo], str]:
    """Get videos based on params.

    Args:
        params (dict): Dictionary of parameters for the API request. The valid params are:
            id (string): Video ID
            user_id (string): User ID
            game_id (string): Game/Category ID
            language (string): ISO 639-1
            period (string): "all", "day", "month", or "week"
            sort (string): "time", "trending", or "views"
            type (string): "all", "archive", "highlight", or "upload"
            first (int): fetch the first n videos
            before (string): reverse pagination
            after (string): forward pagination

    Returns:
        list: list of video info
        string: Pagination cursor
    """
    url = f"{self.API_URL}/videos"
    url = self.__add_parameters(url, params)
    
    r : requests.Response = self.rlrequests.get(url = url, headers = self.DEFAULT_HEADERS)
    
    if r.status_code == 200:
      response_json : dict = r.json()
      response_data : typing.List[dict] = response_json.get("data", [])
      response_pagination : dict = response_json.get("pagination", {})
      
      video_list = [TwitchVideo(data) for data in response_data]
      cursor = response_pagination['cursor'] if "cursor" in response_pagination else ""
      
      return video_list, cursor
    elif r.status_code == 401:
      if self.refresh_authorization():
        return self.get_videos(params)
      else:
        self.__raise_req_error(r)
    else:
      self.__raise_req_error(r)

  def get_all_videos(self, params : dict) -> typing.List[TwitchVideo]:
    """Get all videos based on params (auto-pagination).

    Args:
        params (dict): Dictionary of parameters for the API request. The valid params are:
            id (string): Video ID
            user_id (string): User ID
            game_id (string): Game/Category ID
            language (string): ISO 639-1
            period (string): "all", "day", "month", or "week"
            sort (string): "time", "trending", or "views"
            type (string): "all", "archive", "highlight", or "upload"
            first (int): fetch the first n videos
            before (string): reverse pagination
            after (string): forward pagination

    Returns:
        list: list of videos
    """
    all_videos = []
    while True:
      vids, cursor = self.get_videos(params)

      for vod in vids:
        all_videos.append(vod)
      
      if cursor == "":
        return all_videos
      else:
        params["after"] = cursor

  def get_streams(self, params : dict) -> typing.Tuple[typing.List[TwitchStream], str]:
    """Get a list of streams.

    Args:
        params (dict): Dictionary of parameters for the API request. The valid params are:
            user_id (string): User ID
            user_login (string): Username
            game_id (string): Game/Category ID
            type (string): "all" or "live"
            language (string): ISO 639-1
            first (int): fetch the first n streams
            before (string): reverse pagination
            after (string): forward pagination

    Returns:
        list: list of stream info
        str: Pagination cursor
    """
    url = f"{self.API_URL}/streams/"
    url = self.__add_parameters(url, params)
    
    r : requests.Response = self.rlrequests.get(url = url, headers = self.DEFAULT_HEADERS)
    
    if r.status_code == 200:
      response_json : dict = r.json()
      response_data : typing.List[dict] = response_json.get("data", [])
      response_pagination : dict = response_json.get("pagination", {})
      
      stream_list = [TwitchStream(data) for data in response_data]
      cursor = response_pagination['cursor'] if "cursor" in response_pagination else ""
      
      return stream_list, cursor
    elif r.status_code == 401:
      if self.refresh_authorization():
        return self.get_streams(params)
      else:
        self.__raise_req_error(r)
    else:
      self.__raise_req_error(r)
      
  def get_all_streams(self, params : dict) -> typing.List[TwitchStream]:
    """Get a list of streams.

    Args:
        params (dict): Dictionary of parameters for the API request. The valid params are:
            user_id (string): User ID
            user_login (string): Username
            game_id (string): Game/Category ID
            type (string): "all" or "live"
            language (string): ISO 639-1
            first (int): fetch the first n streams
            before (string): reverse pagination
            after (string): forward pagination

    Returns:
        list: list of stream info
    """
    all_streams = []
    while True:
      streams, cursor = self.get_streams(params)

      for s in streams:
        all_streams.append(s)
      
      if cursor == "":
        return all_streams
      else:
        params["after"] = cursor

  def is_user_live(self, user_id : str) -> bool:
    stream_info = self.get_all_streams({ "user_id": user_id })
    return (len(stream_info) > 0)
  
  def get_channel_emotes(self, broadcaster_id : str) -> typing.Tuple[typing.List[TwitchChannelEmote], str]:
    """Get Channel Emotes

    Args:
        broadcaster_id (string): User ID

    Returns:
        list: List of channel emotes.
        str: Template
    """
    url = f"{self.API_URL}/chat/emotes?broadcaster_id={broadcaster_id}"
    r : requests.Response = self.rlrequests.get(url = url, headers = self.DEFAULT_HEADERS)
    
    if r.status_code == 200:
      response_json : dict = r.json()
      response_data : typing.List[dict] = response_json.get("data", [])
      
      emote_list = [TwitchChannelEmote(data) for data in response_data]
      template : str = response_json.get("template", "")
      
      return emote_list, template
    elif r.status_code == 401:
      if self.refresh_authorization():
        return self.get_channel_emotes(broadcaster_id)
      else:
        self.__raise_req_error(r)
    else:
      self.__raise_req_error(r)
  
  def get_global_emotes(self) -> typing.Tuple[typing.List[TwitchEmote], str]:
    """Get Global Emotes.

    Returns:
        list: List of global emote information.
    """
    url = f"{self.API_URL}/chat/emotes/global"
    r : requests.Response = self.rlrequests.get(url = url, headers = self.DEFAULT_HEADERS)
    
    if r.status_code == 200:
      response_json : dict = r.json()
      response_data : typing.List[dict] = response_json.get("data", [])
      
      emote_list = [TwitchEmote(data) for data in response_data]
      template : str = response_json.get("template", "")
      
      return emote_list, template
    elif r.status_code == 401:
      if self.refresh_authorization():
        return self.get_global_emotes()
      else:
        self.__raise_req_error(r)
    else:
      self.__raise_req_error(r)
    
  def get_custom_rewards(self, params : dict) -> list:
    """Get a list of custom rewards for a channel.

    Args:
        params (dict): dictionary of parameters for the request
          broadcaster_id (int): The ID of the broadcaster whose custom rewards you want to get
          id (string): Rewards ID
          only_manageable_rewards (bool): only get custom rewards that the app may manage

    Returns:
        _type_: _description_
    """
    url = f"{self.API_URL}/channel_points/custom_rewards"
    url = self.__add_parameters(url, params)
    
    r : requests.Response = self.rlrequests.get(url = url, headers = self.DEFAULT_HEADERS)
    
    if r.status_code == 200:
      return r.json()['data']
    elif r.status_code == 401:
      if self.refresh_authorization():
        return self.get_custom_rewards(params)
      else:
        self.__raise_req_error(r)
    else:
      self.__raise_req_error(r)
    
  def create_custom_reward(self, broadcaster_id : str, reward_params : dict) -> dict:
    url = f"{self.API_URL}/channel_points/custom_rewards?broadcaster_id={broadcaster_id}"
    
    r : requests.Response = self.rlrequests.post(url = url, headers = self.DEFAULT_HEADERS, json = reward_params)
    
    if r.status_code == 200:
      return r.json()['data'][0]
    elif r.status_code == 401:
      if self.refresh_authorization():
        return self.create_custom_reward(broadcaster_id, reward_params)
      else:
        self.__raise_req_error(r)
    else:
      self.__raise_req_error(r)
    
  def setup_websocket(self, url_override = None):
    """Connect to the Twitch WebSockets interface for subscribing to events.

    Args:
        url_override (str, optional): Pass a different websockets url for testing purposes. Defaults to None.
    """
    self.TWITCH_WEBSOCKET = TwitchWebSocket(url_override)
    
  def websocket_session_id(self) -> str:
    return self.TWITCH_WEBSOCKET.SESSION_ID
    
  def join_websocket_thread(self):
    self.TWITCH_WEBSOCKET.THREAD.join()
    
  def _add_subscription(self, params : dict) -> bool:
    if not self.TWITCH_WEBSOCKET.CONNECTED:
      print("WebSocket is not connected.")
      return False
    
    url = f'{self.API_URL}/eventsub/subscriptions'
    
    r : requests.Response = self.rlrequests.post(url = url, headers = self.DEFAULT_HEADERS, json=params)
    
    if r.status_code == 202:
      return r.json()['data'][0]['status'] == 'enabled'
    elif r.status_code == 401:
      if self.refresh_authorization():
        return self._add_subscription(params)
      else:
        self.__raise_req_error(r)
    else:
      return False
    
  def get_active_subscriptions(self) -> list:
    """Get active subscriptions in current WebSocket instance.

    Returns:
        list: list of active subscriptions
    """
    url = f'{self.API_URL}/eventsub/subscriptions'
    r : requests.Response = self.rlrequests.get(url = url, headers = self.DEFAULT_HEADERS)
    
    if r.status_code == 200:
      return r.json()['data']
    elif r.status_code == 401:
      if self.refresh_authorization():
        return self.get_active_subscriptions()
      else:
        self.__raise_req_error(r)
    else:
      self.__raise_req_error(r)
    
  def add_subscription(self, event : TwitchEvent, callback) -> bool:
    """Add a subscription to the WebSocket interface.

    Args:
        event (TwitchEvent): Desired subscription event type
        callback (function): Callback for handling notifications matching this subscription

    Returns:
        bool: Success.
    """
    self.TWITCH_WEBSOCKET.add_callback(event.notification_type(), callback)
    return self._add_subscription(event.params())
  
  def subscribe_to_updates(self, user_id : str, callback : typing.Callable[[object, str], None]) -> bool:
    return self.add_subscription(UpdateEvent(user_id, self.websocket_session_id()), callback)
    
  def subscribe_to_follows(self, user_id : str, callback : typing.Callable[[object, str], None]) -> bool:
    return self.add_subscription(FollowEvent(user_id, self.websocket_session_id()), callback)
    
  def subscribe_to_subscriptions(self, user_id : str, callback : typing.Callable[[object, str], None]) -> bool:
    return self.add_subscription(SubscribeEvent(user_id, self.websocket_session_id()), callback)
    
  def subscribe_to_gifted_subscriptions(self, user_id : str, callback : typing.Callable[[object, str], None]) -> bool:
    return self.add_subscription(SubscriptionGiftEvent(user_id, self.websocket_session_id()), callback)
    
  def subscribe_to_subscription_messages(self, user_id : str, callback : typing.Callable[[object, str], None]) -> bool:
    return self.add_subscription(SubscriptionMessageEvent(user_id, self.websocket_session_id()), callback)
  
  def subscribe_to_cheers(self, user_id : str, callback : typing.Callable[[object, str], None]) -> bool:
    return self.add_subscription(CheerEvent(user_id, self.websocket_session_id()), callback)
  
  def subscribe_to_raids(self, user_id : str, callback : typing.Callable[[object, str], None]) -> bool:
    return self.add_subscription(RaidEvent(user_id, self.websocket_session_id()), callback)
  
  def subscribe_to_bans(self, user_id : str, callback : typing.Callable[[object, str], None]) -> bool:
    return self.add_subscription(BanEvent(user_id, self.websocket_session_id()), callback)
  
  def subscribe_to_unbans(self, user_id : str, callback : typing.Callable[[object, str], None]) -> bool:
    return self.add_subscription(UnbanEvent(user_id, self.websocket_session_id()), callback)
  
  def subscribe_to_reward_redemption(self, user_id : str, callback : typing.Callable[[object, str], None], reward_id : str = None) -> bool:
    return self.add_subscription(CustomRewardRedemptionAddEvent(user_id, self.websocket_session_id(), reward_id), callback)
    
  def subscribe_to_stream_online(self, user_id : str, callback : typing.Callable[[object, str], None]) -> bool:
    return self.add_subscription(StreamOnlineEvent(user_id, self.websocket_session_id()), callback)
    
  def subscribe_to_stream_offline(self, user_id : str, callback : typing.Callable[[object, str], None]) -> bool:
    return self.add_subscription(StreamOfflineEvent(user_id, self.websocket_session_id()), callback)


import os
import re
import typing

from .lushrequests import *
from .lushtypes import *

EXT_X_MEDIA_Regex = re.compile(r"#EXT-X-MEDIA:TYPE=([A-Za-z0-9]+),GROUP-ID=\"([A-Za-z0-9]+)\",NAME=\"([A-Za-z0-9]+)\",AUTOSELECT=([A-Za-z0-9]+),DEFAULT=([A-Za-z0-9]+)")
EXT_X_STREAM_INF_Regex = re.compile(r"#EXT-X-STREAM-INF:BANDWIDTH=(\d+),CODECS=\"([A-Za-z0-9\.,]+)\",RESOLUTION=([A-Za-z0-9]+),VIDEO=\"([A-Za-z0-9]+)\"")
    
class TwitchGQL_API:
  API_URL = "https://gql.twitch.tv/gql"
  CLIENT_ID = "kd1unb4b3q4t58fwlpcbzcbnm76a8fp"
  DEFAULT_HEADERS = { "Client-ID": CLIENT_ID }
  REQ = RateLimitedRequests(400, 60)
  
  def __raise_req_error(self, response : requests.Response):
    raise Exception(f'Status {response.status_code}: {response.reason}')
    
  def get_clip(self, clip_id : str) -> GQL_Clip:
    """Get clip information from Twitch's GQL API.

    Args:
        clip_id (string): Clip ID

    Returns:
        dict: Clip info
    """
    content = [
      {
        "operationName": "VideoAccessToken_Clip",
        "variables": {
          "slug": f"{clip_id}"
        },
        "extensions": {
          "persistedQuery": {
            "version": 1,
            "sha256Hash": "36b89d2507fce29e5ca551df756d27c1cfe079e2609642b4390aa4c35796eb11"
          }
        }
      }
    ]
    r : requests.Response = self.REQ.post(url = self.API_URL, headers = self.DEFAULT_HEADERS, json = content)
    
    if r.status_code == 200:
      response_json : typing.List[dict] = r.json()
      if len(response_json) > 0:
        response_data : dict = response_json[0].get("data", {})
        response_clip : dict = response_data.get("clip", {})
        return GQL_Clip(response_clip)
      else:
        raise GQL_Clip.DoesNotExist(f"Clip ID \"{clip_id}\" was not found.")
    else:
      self.__raise_req_error(r)
    
  def get_video(self, video_id):
    """Get video info.

    Args:
        video_id (string): Video ID

    Returns:
        dict: Video information
    """
    content = {
      "query": "query{video(id:\"" + video_id + "\"){id,title,thumbnailURLs(height:180,width:320),createdAt,publishedAt,lengthSeconds,owner{id,displayName}}}",
      "variables": {}
    }
    r : requests.Response = self.REQ.post(url = self.API_URL, headers = self.DEFAULT_HEADERS, json = content)
    
    if r.status_code == 200:
        response_json : dict = r.json()
        response_data : dict = response_json.get("data", {})
        response_video : dict = response_data.get("video", {})
        return GQL_Video(response_video)
    else:
      self.__raise_req_error(r)

  def download_clip(self, clip_id, filename, saveover = False):
    """Download a Twitch clip.

    Args:
        clip_id (string): Clip ID
        filename (string): Filename
        saveover (bool, optional): Overwrite existing files with the same name. Defaults to False.

    Returns:
        bool: Download succeeded/failed.
    """
    if (not saveover and os.path.exists(filename)):
      print(f"Skipping {clip_id} because the filename '{filename}' is taken.")
      return True
    
    info = self.get_clip(clip_id)
    
    url = info.videoQualities[0].sourceURL
    signature = info.playbackAccessToken.signature
    token = info.playbackAccessToken.value
    full_url = f"{url}?sig={signature}&token={token}"
    
    try:
      r = requests.get(full_url)
      with open(filename, 'wb') as outfile:
        outfile.write(r.content)
        return True
    except:
      return False
    
  def get_video_token(self, video_id):
    content = {
      "operationName": "PlaybackAccessToken_Template", 
      "query":"query PlaybackAccessToken_Template($login: String!, $isLive: Boolean!, $vodID: ID!, $isVod: Boolean!, $playerType: String!) {  streamPlaybackAccessToken(channelName: $login, params: {platform: \"web\", playerBackend: \"mediaplayer\", playerType: $playerType}) @include(if: $isLive) {    value    signature    __typename  }  videoPlaybackAccessToken(id: $vodID, params: {platform: \"web\", playerBackend: \"mediaplayer\", playerType: $playerType}) @include(if: $isVod) {    value    signature    __typename  }}",
      "variables": {
        "isLive": False,
        "login": "",
        "isVod": True,
        "vodID": video_id,
        "playerType": "embed"
        }
      }
    
    r = self.REQ.post(url = self.API_URL, headers = self.DEFAULT_HEADERS, json = content)
    
    try:
      return r.json()['data']['videoPlaybackAccessToken']
    except:
      return {}
    
  def get_video_playlist(self, video_id, access_token):
    playlistUrl = f"http://usher.ttvnw.net/vod/{video_id}?nauth={access_token['value']}&nauthsig={access_token['signature']}&allow_audio_only=true&allow_source=true&player=twitchweb"
    
    r : requests.Response = self.REQ.get(url = playlistUrl, headers = self.DEFAULT_HEADERS)
    
    try:
      return str(r.content).split("\\n")
    except:
      return []
    
  def __get_playlist_url(self, video_id, quality : str = "720"):
    token = self.get_video_token(video_id)
    playlist = self.get_video_playlist(video_id, token)
    
    if "vod_manifest_restricted" in playlist[0] or "unauthorized_entitlements" in playlist[0]:
      print("Video restricted. Unable to download.")
      return None
    
    video_qualities : dict[str, dict] = {}
    
    for index in range(len(playlist)):
      item = playlist[index]
      
      media_match = EXT_X_MEDIA_Regex.match(item)
      if media_match:
        string_quality = media_match.group(3)
        
        stream_info_item = playlist[index + 1]
        info_match = EXT_X_STREAM_INF_Regex.match(stream_info_item)
        bandwidth = int(info_match.group(1))
        
        if string_quality not in video_qualities.keys():
          video_qualities[string_quality] = {}
          video_qualities[string_quality]["url"] = playlist[index + 2]
          video_qualities[string_quality]["bandwidth"] = bandwidth
          
    for video_quality in video_qualities.keys():
      if video_quality.lower().startswith(quality.lower()):
        return video_qualities[video_quality]
      
    return video_qualities[list(video_qualities.keys())[0]]
        
    
  def get_video_chapters(self, video_id):
    content = {
      "extensions" : { 
        "persistedQuery": {
          "sha256Hash": "8d2793384aac3773beab5e59bd5d6f585aedb923d292800119e03d40cd0f9b41",
          "version": 1
        }
      },
      "operationName": "VideoPlayer_ChapterSelectButtonVideo",
      "variables": {
        "videoID": video_id
      }
    }
    r = self.REQ.post(url = self.API_URL, headers = self.DEFAULT_HEADERS, json = content)
    try:
      return r.json()['data']['video']['moments']['edges']
    except:
      return []
    
  def __get_or_generate_video_chapters(self, video_id, video_info):
    chapters = self.get_video_chapters(video_id)
    
    if len(chapters) == 0:
      dummy_chapter = {
        "node": {
          "id": "",
          "_type": "GAME_CHANGE",
          "positionMilliseconds": 0,
          "durationMilliseconds": video_info["lengthSeconds"] * 1000,
          "description": "Unknown" if "game" not in video_info else video_info["game"]["displayName"],
          "subDescription": "",
          "details": {
            "game": {
              "id": "-1" if "game" not in video_info else video_info["game"]["id"],
              "displayName": "Unknown" if "game" not in video_info else video_info["game"]["displayName"],
              "boxArtUrl": "" if "game" not in video_info else video_info["game"]["boxArtURL"].replace("{width}", "40").replace("{height}", "40")
            }
          }
        }
      }
      
      chapters.append(dummy_chapter)
      
    return chapters
    
  def __fetch_all_chat(self, video_id):
    """Fetches the chat messages without checking uniqueness or sorting.

    Args:
        video_id (string): Video ID

    Returns:
        list: List of chat messages
    """
    content = [
      {
        "operationName": "VideoCommentsByOffsetOrCursor",
        "variables": {
          "videoID": video_id,
          "contentOffsetSeconds": 0
        },
        "extensions": {
          "persistedQuery": {
            "version": 1,
            "sha256Hash": "b70a3591ff0f4e0313d126c6a1502d79a1c02baebb288227c582044aa76adf6a"
          }
        }
      }
    ]
    
    comments = []
    while True:
      r = self.REQ.post(url = self.API_URL, headers = self.DEFAULT_HEADERS, json = content)
      
      try:
        commentlist = r.json()[0]['data']['video']['comments']['edges']
      except KeyError:
        print("Improper response format.")
        return comments
      
      cursor = None
      for commentdata in commentlist:
        comment = commentdata['node']
        comments.append(comment)
        cursor = commentdata['cursor']
        
      try:
        has_next_page = r.json()[0]['data']['video']['comments']['pageInfo']['hasNextPage']
      except KeyError:
        print("Issue with comment list. Missing pageInfo or hasNextPage.")
        return comments
      
      if has_next_page:
        if "contentOffsetSeconds" in content[0]['variables']:
          content[0]['variables'].pop("contentOffsetSeconds")
        content[0]['variables']['cursor'] = cursor
      else:
        return comments
      
  def get_chat_messages(self, video_id):
    """Get all chat messages and remove duplicates, sort by timestamp

    Args:
        video_id (string): Video ID

    Returns:
        list: List of chat messages
    """
    raw_chat = self.__fetch_all_chat(video_id)
    
    processed_chat = [i for n, i in enumerate(raw_chat) if i not in raw_chat[n + 1:]]
    
    processed_chat.sort(key=lambda a: a['contentOffsetSeconds'])
    return processed_chat
  
  def __get_video_parts_list(self, playlist_url):
    try:
      r = requests.get(playlist_url)
      
      video_chunks : list[str] = str(r.content).split("\\n")
    
      video_list : dict[str, float] = {}
      video_parts : list[str] = []  
    
      for index in range(len(video_chunks)):
        chunk = video_chunks[index]
      
        if chunk.startswith("#EXTINF"):
          nextchunk = video_chunks[index + 1]
          if nextchunk.startswith("#EXT-X-BYTERANGE"):
            nextnextchunk = video_chunks[index + 2]
            if nextnextchunk in video_list.keys():
              video_list[nextnextchunk] += float(chunk[8:-1])
            else:
              video_list[nextnextchunk] = float(chunk[8:-1])
          else:
            video_list[nextchunk] = float(chunk[8:-1])
            
      for video in video_list.keys():
        video_parts.append(video)
        
      return video_parts
    except:
      return []
  
  def download_video(self, video_id : str, filename, quality = "720", saveover = False):
    """Download a Twitch video. Outputs a .ts file that can be converted to your choice of format with ffmpeg.

    Args:
        clip_id (string): Clip ID
        filename (string): Filename
        saveover (bool, optional): Overwrite existing files with the same name. Defaults to False.

    Returns:
        bool: Download succeeded/failed.
    """
    if (not saveover and os.path.exists(filename)):
      print(f"Skipping {video_id} because the filename '{filename}' is taken.")
      return True
    
    playlist_url_bandwidth = self.__get_playlist_url(video_id, quality)
    playlist_url : str = playlist_url_bandwidth["url"]
    
    base_url = playlist_url[:playlist_url.rindex("/") + 1]
    
    video_parts = self.__get_video_parts_list(playlist_url)
  
    with open(filename, "wb") as outfile:
      for part in video_parts:
        full_url = base_url + part
        r = requests.get(full_url)
        outfile.write(r.content)
      outfile.close()